/*
 *		C code editing commands
 *		There are only used when C_MODE is #defined.
 *
 *		Coded by Kazuhiro Takano
 */
/* 90.07.24	Created by K.Takano */

#include	"config.h"	/* 90.12.20  by S.Yoshida */

#ifdef	C_MODE
#include	"def.h"

#ifdef	CHGMISC		/* 97.12.12 by M.Suzuki	*/
#include	"kbd.h"
#endif
#define	GETNUMLEN	6


#ifdef	TABWIDTH	/* 95.08.29 by M.Suzuki	*/
#ifndef	CHGMISC		/* 97.12.12 by M.Suzuki	*/
int defaultTabWidth = 4;
#endif
int tabWidth = 4;
#endif	/* TABWIDTH	*/
static int indent_level			=  2;
			/* Indentation of C statements with respect to	*
			 * containing block.				*/
static int brace_imaginary_offset	=  0;
			/* Imagined indentation of a C open brace that	*
			 * actually follows a statement			*/
static int brace_offset			=  0;		/* not support yet */
			/* Extra indentation for braces, compared with	*
			 * other text in same context.			*/
static int argdecl_indent		=  5;		/* not support yet */
			/* Indentation level of declarations of		*
			 * C function arguments.			*/
static int label_offset			= -2;
			/* Offset of C label lines and case statements	*
			 * relative to usual indentation.		*/
static int continued_statement_offset	=  2;
			/* Extra indent for lines not starting new	*
			 * statements.					*/
static int continued_brace_offset	=  0;		/* not support yet */
			/* Extra indent for substatements that start	*
			 * with open-braces. This is in addition to	*
			 * c-continued-statement-offset.		*/

static int auto_newline			=  FALSE;	/* not support yet */
			/* Non-nil means automatically newline before	*
			 * and after braces, and after colons and	*
			 * semicolons, inserted in C code.		*
			 * Here, FALSE mean nil. (91.01.15 by S.Yoshida) */
static int tab_always_indent		=  TRUE;
			/* Non-nil means TAB in C mode should always	*
			 * reindent the current line, regardless of	*
			 * where in the line point is when the TAB	*
			 * command is used.				*
			 * Here, TRUE mean t. (91.01.15 by S.Yoshida)	*/

/* 91.02.06  Add static declaration for some compiler. by S.Yoshida */
static int	calc_indent ___((void));
static int	adjust_spc ___((int));
static int	check_bal ___((unsigned char*));
static int	count_column ___((LINE*,int));
static int	getnum ___((char*,int*));
#ifdef	CHGMISC	/* 95.06.27 by M.Suzuki	*/
static int SkipBackSpace ___((LINE*,int));
static int SkipSpace ___((LINE*));
#endif	/* CHGMISC	*/


static int
SkipBackSpace(lp,pos)
LINE* lp;
int pos;
{
	for (pos--; pos >= 0; pos--) {
		char c;
		c = lgetc(lp, pos);
		if (c != ' ' && c != '\t')
		break;
	}
	return(pos);
}

static int
SkipSpace(lp)
LINE* lp;
{
	int pos;

	for (pos = 0; pos < llength(lp); pos++) {
		unsigned char c;
		c = lgetc(lp, pos);
		if (c != ' ' && c != '\t')
			break;
	}
	return(pos);
}

static int
nth_bslash(s,n)
char* s;
int n;
{
	int c;

	for(c=0;n;n--,c++){
		if( s[n] != '\\' ){
			break;
		}
	}
	return c & 1;
}

#define	nth_kanji(s,n)	ISKANJI(s[n])

static int
calc_indent ___((void))
{
	register LINE	*lp;
	register int	bo;
	unsigned char	bal1, bal2, c;
	LINE	*tlp;
	int	tbo, cbo;
	int 	s;
	int	depth;
#ifndef	BUGFIX1		/* 93.03.21 by M.Suzuki	*/
	int	in_paren = FALSE;
	int	in_comment = FALSE;
	int	has_semi = FALSE;
#endif
	int	null_line = TRUE;
	int	firstcheck = TRUE;
#ifndef	BUGFIX1		/* 92.12.16 by M.Suzuki	*/
	int	in_strings = FALSE;	/* 91.01.13  by S.Yoshida */
	int	numesc = 0;		/* 91.01.13  by S.Yoshida */
#endif
	unsigned char	termchar = 0;		/* 91.01.13  by S.Yoshida */
	int	indent = 0;
#ifdef	BUGFIX1		/* 92.12.16 by M.Suzuki	*/
	int kflag;
	int qflag;
#endif

	lp = curwp->w_dotp;
	c = lgetc(lp,(cbo=SkipSpace(lp)));
#ifdef	CHGMISC	/* 93.03.21 by M.Suzuki	*/
	if (c == '#')			/* for pre-processer line	*/
		return (0);		/* no indent			*/
#endif

#ifdef	BUGFIX1		/* 94.12.14 by M.Suzuki	*/
	bal1 = depth = termchar = '\0';
	kflag = 0;
	qflag = 0;
	for (bo = cbo; bo < llength(lp); bo++) {
		c = lgetc(lp, bo);
		if( ISKANJI(c) || kflag ){
			kflag = ~kflag;
		} else
		if( termchar ){
			if( c == '\\' ){
				kflag = ~kflag;
			} else
			if ((unsigned char)c == termchar){				/* End of strings.	*/
				termchar = '\0';
			}
		} else
		if( c == '?'){
			qflag = 1;
		} else
		if (c == ':' && !qflag ) {
			indent += label_offset;
			break;
		} else
		if ( c == '\"' || c == '\'') {		/* Not in_strings. */
			termchar = c;
		}
	}
#else
	for (bo = cbo; bo < llength(lp); bo++) {
		/* 91.01.13  Modify to consider strings  by S.Yoshida */
		c = lgetc(lp, bo);
		if (c == ':' && !in_strings) {
			indent += label_offset;
			break;
		} else
		if (in_strings) {
			if (c == '\"' || c == '\'') {
				if (c == termchar && (numesc % 2) == 0) {
					in_strings = FALSE;
				} else {
					numesc = 0;
				}
			} else
			if (c == '\\') {
				numesc++;
			} else {
				numesc = 0;
			}
		} else {	/* Not in_strings. */
			if (c == '\"' || c == '\'') {
				in_strings = TRUE;
				termchar = c;
				numesc = 0;
			}
		}
	}
#endif	/* BUGFIX1	*/

#ifdef	BUGFIX1		/* 92.12.16 by M.Suzuki	*/
	bo = 0;
	termchar = '\0';	/* 91.01.13  by S.Yoshida */
	{
		int	in_paren;
		int	in_comment;
		int	has_semi;

		has_semi = FALSE;
		in_paren = 0;
		in_comment = 0;
		for (;;) {
			if (bo == 0) {			/* beginning of line	*/
				if (firstcheck == TRUE && null_line == FALSE) {
					if (has_semi == FALSE){
						indent += continued_statement_offset;
					} else {
						firstcheck = FALSE;
					}
				}
				null_line = TRUE;
				has_semi = FALSE;

				lp = lback(lp);
				if (lp == curbp->b_linep)
					break;		/* beginning of buffer	*/
				if ((bo = llength(lp)) == 0)
					continue;	/* skip null line	*/
			}
			bo--;
			c = lgetc(lp, bo);		/* somewhere in middle	*/

			if( bo ){
				if( nth_kanji(lp->l_text,bo-1) ){	/* kanji 2nd check	*/
					bo--;			/* kanji 1st skip	*/
					continue;
				}
				if( nth_bslash(lp->l_text,bo-1) ){
					bo--;
					continue;
				}
			} else
			if( c == '#' ){
				null_line = TRUE;
				has_semi = FALSE;
				continue;
			}

			if (c == ' ' || c == '\t'){
				continue;		/* skip space		*/
			}

			if( bo ){
				if (c == '*' && lgetc(lp, bo - 1) == '/') {
						/* comment start mark	*/
					if (in_comment) {
						bo--;
						in_comment = 0;
						continue;
					}
					return (count_column(lp, bo));
				}
			}
			if (in_comment){
				continue;		/* skip comment		*/
			}
			if( bo ){
				if (c == '/' && lgetc(lp, bo - 1) == '*') {
					/* comment end mark	*/
					bo--;
					in_comment = ~in_comment;
					continue;
				}
			}
			null_line = FALSE;

			if((unsigned char)c == termchar){
/*				if( !bo || lgetc(lp,bo-1) != '\\'){	*/
					termchar = '\0';
/*				}	*/
				continue;
			}
			if( termchar ){
				continue;
			}
			if( c == '\'' || c == '\"' ){
				termchar = c;
				continue;
			}

			if (in_paren) {
				if (bal2 == '{')
					has_semi = TRUE;
				if (c == bal2) {	/* left parenthesis	*/
					if (--depth == 0)
						in_paren = 0;
				}
				else if (c == bal1) {	/* right parenthesis	*/
					depth++;
				}
				continue;
			}


			if (c == ';' || c == ':') {
				has_semi = TRUE;
				continue;
			}

			bal1 = bal2 = c;
			s = check_bal(&bal2);
			if (s < 0) {			/* right parenthesis	*/
				if (c == '}')
					has_semi = TRUE;
				depth = 1;
				in_paren = ~in_paren;	/* search match paren	*/
				continue;
			}
			if (s > 0) {			/* left parenthesis	*/
				break;
			}
		}	/* end of for	*/
	}
#else
	bo = 0;
	in_strings = FALSE;	/* 91.01.13  by S.Yoshida */
	numesc = 0;		/* 91.01.13  by S.Yoshida */
	termchar = '\0';	/* 91.01.13  by S.Yoshida */
	for (;;) {
		if (bo == 0) {			/* beginning of line	*/
			if (firstcheck == TRUE && null_line == FALSE) {
				if (has_semi == FALSE)
					indent += continued_statement_offset;
				else
					firstcheck = FALSE;
			}
			null_line = TRUE;
			has_semi = FALSE;

			lp = lback(lp);
			if (lp == curbp->b_linep)
				break;		/* beginning of buffer	*/
			if ((bo = llength(lp)) == 0)
				continue;	/* skip null line	*/
		}
		bo--;
		c = lgetc(lp, bo);		/* somewhere in middle	*/

		if (c == ' ' || c == '\t')
			continue;		/* skip space		*/

		if (c == '*' && bo > 0 && lgetc(lp, bo - 1) == '/') {
						/* comment start mark	*/
			if (in_comment == TRUE) {
				bo--;
				in_comment = FALSE;
				continue;
			}
			return (count_column(lp, bo));
		}
		if (in_comment == TRUE)
			continue;		/* skip comment		*/
		if (c == '/' && bo > 0 && lgetc(lp, bo - 1) == '*') {
						/* comment end mark	*/
			bo--;
			in_comment = TRUE;
			continue;
		}
		null_line = FALSE;

		/* 91.01.13  Add to consider strings.  by S.Yoshida */
		if (in_strings) {
			if (c == '\"' || c == '\'') {
				if (c == termchar && (numesc % 2) == 0) {
					in_strings = FALSE;
				} else {
					numesc = 0;
				}
			} else if (c == '\\') {
				numesc++;
			} else {
				numesc = 0;
			}
			continue;
		} else {	/* Not in strings. */
			if (c == '\"' || c == '\'') {
				in_strings = TRUE;
				termchar = c;
				numesc = 0;
			}
		}

		if (in_paren == TRUE) {
			if (bal2 == '{')
				has_semi = TRUE;
			if (c == bal2) {	/* left parenthesis	*/
				if (--depth == 0)
					in_paren = FALSE;
			}
			else if (c == bal1) {	/* right parenthesis	*/
				depth++;
			}
			continue;
		}

		if (c == ';' || c == ':') {
			has_semi = TRUE;
			continue;
		}

		bal1 = bal2 = c;
		s = check_bal(&bal2);
		if (s < 0) {			/* right parenthesis	*/
			if (c == '}')
				has_semi = TRUE;
			depth = 1;
			in_paren = TRUE;	/* search match paren	*/
			continue;
		}
		if (s > 0) {			/* left parenthesis	*/
			break;
		}
	}
#endif	/* BUGFIX1	*/
	tlp = lp;
	tbo = bo;


	if (lp == curbp->b_linep) {		/* line is at top level	*/
		return (-1);
	}

	if (bal1 != '{')			/* line is expression	*/
		return (count_column(tlp, tbo) + 1);


	bo = SkipBackSpace(lp,bo);
	if (++bo > 0)
		indent += brace_imaginary_offset;


	s = 0;
	for (bo = 0; bo < llength(tlp); bo++) {
		c = lgetc(tlp, bo);
		if (c != ' ' && c != '\t')
			break;
		if (c == '\t')
#ifdef	TABWIDTH	/* 95.08.29 by M.Suzuki	*/
			s += TabOffset(s);
#else
			s |= 0x07;
#endif
		s++;
	}

#ifdef	BUGFIX1		/* 94.12.14 by M.Suzuki	*/
	termchar = '\0';	/* 91.01.13  by S.Yoshida */
	kflag = 0;
	lp = curwp->w_dotp;
	for (bo = cbo; bo < llength(lp); bo++) {
		depth = 0;
		c = lgetc(lp, bo);
		if( ISKANJI(c) || kflag ){
			kflag = ~kflag;
		} else
		if( termchar ){
			if( c == '\\' ){
				kflag = ~kflag;
			} else
			if (termchar == (unsigned char)c) {
				termchar = '\0';
			}
			continue;
		} else
		if (c == '\"' || c == '\'') {	/* Not in_strings. */
			termchar = c;
		} else
		if (c == bal2) {
			if (depth == 0) {
				return (s + indent);
					/* include close parenthesis	*/
			}
			depth--;
		} else
		if (c == bal1) {
			depth++;
		}
	}
#else
	in_strings = FALSE;	/* 91.01.13  by S.Yoshida */
	numesc = 0;		/* 91.01.13  by S.Yoshida */
	termchar = '\0';	/* 91.01.13  by S.Yoshida */
	lp = curwp->w_dotp;
	for (bo = cbo; bo < llength(lp); bo++) {
		depth = 0;
		c = lgetc(lp, bo);
		/* 91.01.13  Modify & add to consider strings.  by S.Yoshida */
		if (c == bal2 && !in_strings) {
			if (depth == 0) {
				return (s + indent);
					/* include close parenthesis	*/
			}
			depth--;
		} else if (c == bal1 && !in_strings) {
			depth++;
		} else if (in_strings) {
			if (c == '\"' || c == '\'') {
				if (c == termchar && (numesc % 2) == 0) {
					in_strings = FALSE;
				} else {
					numesc = 0;
				}
			} else if (c == '\\') {
				numesc++;
			} else {
				numesc = 0;
			}
		} else {	/* Not in_strings. */
			if (c == '\"' || c == '\'') {
				in_strings = TRUE;
				termchar = c;
				numesc = 0;
			}
		}
	}
#endif	/* BUGFIX1	*/

	return (s + indent_level + indent);
}


static int
adjust_spc(nicol)
int nicol;
{
	register int	i;
	int	cbo;

	cbo = curwp->w_doto;
	i = SkipSpace(curwp->w_dotp);
	if ((cbo -= i) < 0)
		cbo = 0;
	curwp->w_doto = 0;		/* delete heading space */
	if (i != 0 && ldelete((RSIZE)i, KNONE) == FALSE)
		return (FALSE);

	if (				/* insert space (and/or tab) */
#ifdef	NOTAB
		(curbp->b_flag & BFNOTAB) ?
	      (cbo += nicol, linsert(nicol, ' ') == FALSE) :
#endif
#ifdef	TABWIDTH	/* 95.08.29 by M.Suzuki	*/
	    ((i = nicol / tabWidth) != 0 && (cbo += i, linsert(i, '\t') == FALSE)) ||
	    ((i = nicol % tabWidth) != 0 && (cbo += i, linsert(i, ' ' ) == FALSE)))
#else
		((i = nicol / 8) != 0 && (cbo += i, linsert(i, '\t') == FALSE)) ||
		((i = nicol % 8) != 0 && (cbo += i, linsert(i, ' ' ) == FALSE)))
#endif
		return (FALSE);

	curwp->w_doto = cbo;
	return (TRUE);
}


static int
check_bal(balc)
unsigned char *balc;
{
	static struct {
		unsigned char left, right;
	} bal[] = {
		{ '(', ')' },
		{ '[', ']' },
		{ '{', '}' },
		{ '\0','\0'}
	};
	register int	i;

	for (i = 0; bal[i].right != '\0'; i++) {
		if (bal[i].right == *balc) {
			*balc = bal[i].left;
			return (-1);
		}
		if (bal[i].left == *balc) {
			*balc = bal[i].right;
			return (1);
		}
	}

	return (0);
}


static int
count_column(lp, bo)
LINE* lp;
int bo;
{
	register int i;
	register int col;

	col = 0;
	for (i = 0; i < bo; i++) {
		if (lgetc(lp, i) == '\t')
#ifdef	TABWIDTH	/* 95.08.29 by M.Suzuki	*/
			col += TabOffset(col);
#else
			col |= 7;
#endif
		col++;
	}
	return (col);
}

/*
 * COMMAND: electric-c-brace
 */
/*ARGSUSED*/
int
cm_brace(f, n)
int f;
int n;
{
	if (warnreadonly()) {	/* If this buffer is read-only, */
							/* do only displaying warning.	*/
		return TRUE;
	}

	if (n < 0)
		return (FALSE);

	if (selfinsert(FFRAND, n) == FALSE)
		return (FALSE);
	return (cm_indent(FFRAND, 1));
}


/*
 * COMMAND: electric-c-semi
 */
/*ARGSUSED*/
int
cm_semi(f, n)
int f;
int n;
{
	if (warnreadonly()) {	/* If this buffer is read-only, */
							/* do only displaying warning.	*/
		return TRUE;
	}

	if (auto_newline)
		return (cm_term(FFRAND, n));
	else
		return (selfinsert(FFRAND, n));
}


/*
 * COMMAND: electric-c-terminator
 */
/*ARGSUSED*/
int
cm_term(f, n)
int f;
int n;
{
	if (warnreadonly()) {	/* If this buffer is read-only, */
							/* do only displaying warning.	*/
		return TRUE;
	}

	if (selfinsert(FFRAND, n) == FALSE)
		return (FALSE);

	return (cm_indent(FFRAND, n));
}

/*
 * COMMAND: c-newline-and-indent
 */
/*ARGSUSED*/
int
cm_lfindent(f, n)
int f;
int n;
{
	if (warnreadonly()) {	/* If this buffer is read-only, */
							/* do only displaying warning.	*/
		return TRUE;
	}

	if (n < 0)
		return (FALSE);

	curwp->w_doto = SkipBackSpace(curwp->w_dotp,curwp->w_doto) + 1;

	if (newline(FFRAND, n) == FALSE)
		return (FALSE);
	return (cm_indent(FFRAND, 1));
}

/*
 * COMMAND: c-indent-command
 */
/*ARGSUSED*/
int
cm_indent(f, n)
int f;
int n;
{
	int i;

	if (warnreadonly()) {	/* If this buffer is read-only, */
							/* do only displaying warning.	*/
		return TRUE;
	}

	if ((f & (FFRAND | FFARG)) == 0 && tab_always_indent == FALSE) {
		if (SkipBackSpace(curwp->w_dotp,curwp->w_doto) >= 0)
			return (selfinsert(FFRAND, n));
	}

	i = calc_indent();
	if (i < 0)
		i = 0;
	return (adjust_spc(i));
}


/*
 *	Set and Display variables
 */

/*
 * COMMAND: set-c-indent-level
 */
/*ARGSUSED*/
int
cm_set_indent(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-indent-level", &n) == FALSE)
			return (FALSE);
	}
	indent_level = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-brace-imaginary-offset
 */
/*ARGSUSED*/
int
cm_set_imagin(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-brace_imaginary-offset", &n) == FALSE)
			return (FALSE);
	}
	brace_imaginary_offset = n;
	return (TRUE);
}



/*
 * COMMAND: set-c-brace-offset
 */
/*ARGSUSED*/
int
cm_set_brace(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-brace-offset", &n) == FALSE)
			return (FALSE);
	}
	brace_offset = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-argdecl-indent
 */
/*ARGSUSED*/
int
cm_set_arg(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-argdecl-indent", &n) == FALSE)
			return (FALSE);
	}
	argdecl_indent = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-label-offset
 */
/*ARGSUSED*/
int
cm_set_label(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-label-offset", &n) == FALSE)
			return (FALSE);
	}
	label_offset = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-continued-statement-offset
 */
/*ARGSUSED*/
int
cm_set_cstat(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-continued-statement-offset", &n) == FALSE)
			return (FALSE);
	}
	continued_statement_offset = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-continued-brace-offset
 */
/*ARGSUSED*/
int
cm_set_cbrace(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("c-continued-brace-offset", &n) == FALSE)
			return (FALSE);
	}
	continued_brace_offset = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-auto-newline
 * 91.01.15  Modified by S.Yoshida
 * Add routine to consider "t" and "nil" argument.
 */
/*ARGSUSED*/
int
cm_set_newl(f, n)
int f;
int n;
{
	register int	s;
	char	buf[NFILEN];

	if ((f & FFARG) == 0) {
		if ((s = ereply("c-auto-newline : ", buf, NFILEN)) != TRUE)
			return (s);
		if (ISDIGIT(buf[0]) || buf[0] == '-')
			n = (atoi(buf) > 0);
		else if (buf[0] == 't' || buf[0] == 'T')
			n = TRUE;
		else /* if (buf[0] == 'n' || buf[0] == 'N') */
			n = FALSE;
	}
	auto_newline = n;
	return (TRUE);
}


/*
 * COMMAND: set-c-tab-always-indent
 * 91.01.15  Modified by S.Yoshida
 * Add routine to consider "t" and "nil" argument.
 */
/*ARGSUSED*/
int
cm_set_tab(f, n)
int f;
int n;
{
	register int	s;
	char	buf[NFILEN];

	if ((f & FFARG) == 0) {
		if ((s = ereply("c-tab-always-indent : ", buf, NFILEN)) != TRUE)
			return (s);
		if (ISDIGIT(buf[0]) || buf[0] == '-')
			n = (atoi(buf) > 0);
		else if (buf[0] == 't' || buf[0] == 'T')
			n = TRUE;
		else /* if (buf[0] == 'n' || buf[0] == 'N') */
			n = FALSE;
	}
	tab_always_indent = n;
	return (TRUE);
}

#ifdef	TABWIDTH	/* 95.08.29 by M.Suzuki	*/
/*
 * COMMAND: set-default-tab-width
 */
/*ARGSUSED*/
int
default_tab_width(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("set-default-tab-width", &n) == FALSE)
			return (FALSE);
	}
	if( n == 0 ){
		n = 1;
	}
#ifdef	CHGMISC		/* 97.12.12 by M.Suzuki	*/
	defb.tab_width = n;
#else
	defaultTabWidth = n;
#endif
	tabWidth = n;
  	curwp->w_flag |= WFHARD;
	return (TRUE);
}

/*
 * COMMAND: tab-width
 */
/*ARGSUSED*/
int
tab_width(f, n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("tab-width", &n) == FALSE)
			return (FALSE);
	}
	if( n == 0 ){
		n = 1;
	}
	curbp->b_tabwidth = n;
  	curwp->w_flag |= WFHARD;
	return (TRUE);
}
#endif	/* TABWIDTH	*/


/*
 * COMMAND: list-c-mode-variables
 */
/*ARGSUSED*/
int
cm_list_var(f, n)
int f;
int n;
{
	register BUFFER	*bp;
	register WINDOW	*wp;
	char	 line[80];

	if ((bp = bfind("*C Mode Variables*", "")) == NULL) return FALSE;
	bp->b_flag &= ~BFCHG;		/* Blow away old.	*/
 	if (bclear(bp) != TRUE) return FALSE;

	if (addline(bp, "* List of variables controlling indentation style *") == FALSE) return FALSE;

	sprintf(line, "\tc-indent-level : %d", indent_level);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-brace-imaginary-offset : %d", brace_imaginary_offset);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-brace-offset : %d", brace_offset);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-argdecl-indent : %d", argdecl_indent);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-label-offset : %d", label_offset);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-continued-statement-offset : %d", continued_statement_offset);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-continued-brace-offset : %d", continued_brace_offset);
	if (addline(bp, line) == FALSE) return FALSE;

	sprintf(line, "\tc-auto-newline : %s",
		auto_newline ? "T" : "NIL");
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tc-tab-always-indent : %s",
		tab_always_indent ? "T" : "NIL");
	if (addline(bp, line) == FALSE) return FALSE;

	if ((wp = popbuf(bp)) == NULL)	return FALSE;
	bp->b_dotp = lforw(bp->b_linep); /* put dot at beginning of buffer */
	bp->b_doto = 0;
	wp->w_dotp = bp->b_dotp;	/* fix up if window already on screen */
	wp->w_doto = bp->b_doto;
	return TRUE;
}

#ifdef	AUTOSAVE	/* 96.12.25 by M.Suzuki	*/
#include <unistd.h>


int
auto_save(f,n)
int f;
int n;
{
	if ((f & FFARG) == 0) {
		if (getnum("auto-save-interval", &n) == FALSE)
			return (FALSE);
	}
	autoSaveInterval = n;
	alarm(autoSaveInterval);
	return (TRUE);
}
#endif	/* AUTOSAVE	*/

static int
getnum(prompt, num)
char *prompt;
int *num;
{
	char numstr[GETNUMLEN];

#ifdef	BUGFIX1		/* 97.11.10 by M.Suzuki	*/
	if (ereply("%s : ", numstr, GETNUMLEN, prompt) == !TRUE)
#else
	if (ereply("%s : ", numstr, GETNUMLEN, prompt) == FALSE)
#endif	/* BUGFIX1	*/
		return (FALSE);
	*num = atoi(numstr);
	return (TRUE);
}

#endif	/* C_MODE */

