/*
 *		Terminal independent keyboard handling.
 */
/* 90.01.29	Modified for Ng 1.0 by S.Yoshida */

#include	"config.h"	/* 90.12.20  by S.Yoshida */
#include	"def.h"
#include	"kbd.h"

#define EXTERN
#include	"key.h"

#ifndef NO_MACRO
#include "macro.h"
#endif

#ifdef	DO_METAKEY
#ifndef METABIT
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
#define METABIT 0x100
#else	/* NOT KANJI */
#define METABIT 0x80
#endif	/* KANJI */
#endif	/* METABIT */

static int use_metakey = TRUE;

/*
 * Toggle the value of use_metakey
 */
int
do_meta(f, n)
int f;
int n;
{
	if(f & FFARG)	use_metakey = n > 0;
	else		use_metakey = !use_metakey;
	ewprintf("Meta keys %sabled", use_metakey ? "en" : "dis");
	return TRUE;
}
#endif

#ifdef	BSMAP
static int bs_map = BSMAP;
/*
 * Toggle backspace mapping
 */
int
bsmap(f, n)
int f;
int n;
{
	if(f & FFARG)	bs_map = n > 0;
	else		bs_map = ! bs_map;
	ewprintf("Backspace mapping %sabled", bs_map ? "en" : "dis");
	return TRUE;
}
#endif

#ifndef NO_DPROMPT
#define PROMPTL 80
  char	prompt[PROMPTL], *promptp;
#endif

static	int	pushed = FALSE;
static	int	pushedc;

VOID
ungetkey(c)
int	c;
{
#ifdef	DO_METAKEY
	if(use_metakey && pushed && c==CCHR('[')) pushedc |= METABIT;
	else
#endif
		pushedc = c;
	pushed = TRUE;
}

int
getkey(flag)
int	flag;
{
	int	c;

#ifndef NO_DPROMPT
	if(flag && !pushed) {
		if(prompt[0]!='\0' && ttwait()) {
			ewprintf("%s", prompt);	/* avoid problems with % */
			update();		/* put the cursor back	 */
			epresf = KPROMPT;
		}
		if(promptp > prompt) *(promptp-1) = ' ';
	}
#endif
#ifdef	AUTOSAVE	/* 96.12.24 by M.Suzuki	*/
	AutoSaveUnlock();
#endif	/* AUTOSAVE	*/
	if(pushed) {
		c = pushedc;
		pushed = FALSE;
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
	} else	c = kgetkey();
#else	/* NOT KANJI */
	} else	c = getkbd();
#endif	/* KANJI */
#ifdef 	BSMAP
	if(bs_map)
		if(c==CCHR('H')) c=CCHR('?');
		else if(c==CCHR('?')) c=CCHR('H');
#endif	
#ifdef	DO_METAKEY
	if(use_metakey && (c&METABIT)) {
		pushedc = c & ~METABIT;
		pushed = TRUE;
		c = CCHR('[');
	}
#endif
#ifndef NO_DPROMPT
	if(flag && promptp < &prompt[PROMPTL - 5]) {
	    promptp = keyname(promptp, c);
	    *promptp++ = '-';
	    *promptp = '\0';
	}
#endif
#ifdef	AUTOSAVE	/* 96.12.24 by M.Suzuki	*/
	AutoSaveLock();
#endif	/* AUTOSAVE	*/
	return c;
}

/*
 * doscan scans a keymap for a keyboard character and returns a pointer
 * to the function associated with that character.  Sets ele to the
 * keymap element the keyboard was found in as a side effect.
 */

MAP_ELEMENT *ele;

PF
doscan(map, c)
register KEYMAP *map;
register int	c;
{
    register MAP_ELEMENT *elec = &map->map_element[0];	/* local register copy for faster access */
    register MAP_ELEMENT *last = &map->map_element[map->map_num];

    while(elec < last && c > elec->k_num) elec++;
    ele = elec;			/* used by prefix and binding code	*/
    if(elec >= last || c < elec->k_base)
	return map->map_default;
    return elec->k_funcp[c - elec->k_base];
}

int
doin ___((void))
{
	KEYMAP	*curmap;
	PF	funct;
#ifdef CANNA
	int d;
#endif

#ifndef NO_DPROMPT
    *(promptp = prompt) = '\0';
#endif
#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
    fepmode_on();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
	canna_on();
#endif
    curmap = curbp->b_modes[curbp->b_nmodes]->p_map;
    key.k_count = 0;

    while((funct=doscan(curmap,(key.k_chars[key.k_count++]=getkey(TRUE))))
		== prefix) {
#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
	fepmode_off();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
	canna_off();
#endif
	curmap = ele->k_prefmap;
    }
#ifndef NO_MACRO
    if(macrodef && macrocount < MAXMACRO)
	macro[macrocount++].m_funct = funct;
#endif
    return (*funct)(0, 1);
}

int
rescan(f, n)
int f;
int n;
{
    int c;
    register KEYMAP *curmap;
    int i;
    PF	fp;
    int mode = curbp->b_nmodes;

	for(;;) {
		if(ISUPPER(key.k_chars[key.k_count-1])) {
			c = TOLOWER(key.k_chars[key.k_count-1]);
			curmap = curbp->b_modes[mode]->p_map;
			for(i=0; i < key.k_count-1; i++) {
				if((fp=doscan(curmap,(key.k_chars[i]))) != prefix) break;
				curmap = ele->k_prefmap;
			}
			if(fp==prefix) {
				if((fp = doscan(curmap, c)) == prefix) {
#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
					fepmode_off();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
					canna_off();
#endif
				    while((fp=doscan(curmap,key.k_chars[key.k_count++] =
									 getkey(TRUE))) == prefix)
						curmap = ele->k_prefmap;
				}
				if(fp!=rescan) {
#ifndef NO_MACRO
					if(macrodef && macrocount <= MAXMACRO)
						macro[macrocount-1].m_funct = fp;
#endif
				    return (*fp)(f, n);
				}
		    }
		}
	/* try previous mode */
#ifdef KANJI
		if(--mode < 0) {
		    if (ISKANJI(key.k_chars[key.k_count-1]))
				getkey(FALSE);
		    return ABORT;
		}
#else
		if(--mode < 0) return ABORT;
#endif
		curmap = curbp->b_modes[mode]->p_map;
		for(i=0; i < key.k_count; i++) {
		    if((fp=doscan(curmap,(key.k_chars[i]))) != prefix) break;
		    curmap = ele->k_prefmap;
		}
		if(fp==prefix) {
#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
		    fepmode_off();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
			canna_off();
#endif
		    while((fp=doscan(curmap,key.k_chars[i++]=getkey(TRUE)))
				  == prefix)
				curmap = ele->k_prefmap;
		    key.k_count = i;
		}
		if(fp!=rescan && i>=key.k_count-1) {
#ifndef NO_MACRO
		    if(macrodef && macrocount <= MAXMACRO)
				macro[macrocount-1].m_funct = fp;
#endif
		    return (*fp)(f, n);
		}
    }
}

int
universal_argument(f, n)
int f;
int n;
{
    int c, nn=4;
    KEYMAP *curmap;
    PF	funct;

#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
    fepmode_off();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
	canna_off();
#endif
    if(f&FFUNIV) nn *= n;
    for(;;) {
	key.k_chars[0] = c = getkey(TRUE);
	key.k_count = 1;
	if(c == '-') return negative_argument(f, nn);
	if( ISDIGIT(c) ) return digit_argument(f, nn);
	curmap = curbp->b_modes[curbp->b_nmodes]->p_map;
	while((funct=doscan(curmap,c)) == prefix) {
	    curmap = ele->k_prefmap;
	    key.k_chars[key.k_count++] = c = getkey(TRUE);
	}
	if(funct != universal_argument) {
#ifndef NO_MACRO
	    if(macrodef && macrocount < MAXMACRO-1) {
		if(f&FFARG) macrocount--;
		macro[macrocount++].m_count = nn;
		macro[macrocount++].m_funct = funct;
	    }
#endif
	    return (*funct)(FFUNIV, nn);
	}
	nn <<= 2;
    }
}

/*ARGSUSED*/
int
digit_argument(f, n)
int f;
int n;
{
    int nn, c;
    KEYMAP *curmap;
    PF	funct;

#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
    fepmode_off();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
	canna_off();
#endif
    nn = key.k_chars[key.k_count-1] - '0';
    for(;;) {
	c = getkey(TRUE);
	if( !ISDIGIT(c) ) break;
	nn *= 10;
	nn += c - '0';
    }
    key.k_chars[0] = c;
    key.k_count = 1;
    curmap = curbp->b_modes[curbp->b_nmodes]->p_map;
    while((funct=doscan(curmap,c)) == prefix) {
	curmap = ele->k_prefmap;
	key.k_chars[key.k_count++] = c = getkey(TRUE);
    }
#ifndef NO_MACRO
    if(macrodef && macrocount < MAXMACRO-1) {
	if(f&FFARG) macrocount--;
	else macro[macrocount-1].m_funct = universal_argument;
	macro[macrocount++].m_count = nn;
	macro[macrocount++].m_funct = funct;
    }
#endif
    return (*funct)(FFOTHARG, nn);
}

int
negative_argument(f, n)
int f;
int n;
{
    int nn = 0, c;
    KEYMAP *curmap;
    PF	funct;

#ifdef FEPCTRL	/* 90.11.26  by K.Takano */
    fepmode_off();
#endif
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
    canna_off();
#endif
    for(;;) {
	c = getkey(TRUE);
	if( !ISDIGIT(c) ) break;
	nn *= 10;
	nn += c - '0';
    }
    if(nn) nn = -nn;
    else nn = -n;
    key.k_chars[0] = c;
    key.k_count = 1;
    curmap = curbp->b_modes[curbp->b_nmodes]->p_map;
    while((funct=doscan(curmap,c)) == prefix) {
	curmap = ele->k_prefmap;
	key.k_chars[key.k_count++] = c = getkey(TRUE);
    }
#ifndef NO_MACRO
    if(macrodef && macrocount < MAXMACRO-1) {
	if(f&FFARG) macrocount--;
	else macro[macrocount-1].m_funct = universal_argument;
	macro[macrocount++].m_count = nn;
	macro[macrocount++].m_funct = funct;
    }
#endif
    return (*funct)(FFNEGARG, nn);
}

/*
 * Insert a character.	While defining a macro, create a "LINE" containing
 * all inserted characters.
 */
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
int	no_k2nd = FALSE;		/* We have no KANJI 2nd byte	*/
int	inkfill = FALSE;		/* Now we are in a fillword().	*/
#endif	/* KANJI */

int
selfinsert(f, n)
int f;
int n;
{
    register int c;
    int count;
#ifndef NO_MACRO
    LINE *lp;
#endif
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
    int	lkanji2nd;			/* For over write mode.	*/
#endif	/* KANJI */

    if (n < 0)	return FALSE;
    if (n == 0) return TRUE;
    c = key.k_chars[key.k_count-1];
	if (warnreadonly()) {	/* If this buffer is read-only, */
							/* do only displaying warning.	*/
		return TRUE;
	}
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
    /* In a autofill mode, normally fill trigger is only ' ' char. But	*/
    /* KANJI version require any KANJI char is also fill trigger.	*/
    /* This must be done at keymap.c to add KANJI fill trigger list in	*/
    /* a fillmap. But there are too many KANJI chars, so we use this	*/
    /* easy way.							*/
    if((curbp->b_flag & BFAUTOFILL) && 		/* Autofill mode and	*/
       !inkfill && no_k2nd) {			/* KANJI 2nd byte.	*/
	    int	s;
	    inkfill = TRUE;
	    s = fillword(f, n);		/* fill word with KANJI char.	*/
	    inkfill = FALSE;
	    return (s);
    }					/* End of autofill mode add routine. */
    if (no_k2nd) {			/* If there is only KANJI 1st byte, */
	    no_k2nd = FALSE;		/* we believe 'c' is KANJI 2nd byte.*/
    } else if (ISKANJI(c)) {
	    if ((n % 2) == 0) {		/* This is easy bug fix. */
		    n |= 0x01;
	    }
	    no_k2nd = TRUE;		/* When there is no KANJI 2nd	*/
					/* byte, we don't do update().	*/
    }
#endif	/* KANJI */
#ifndef NO_MACRO
    if(macrodef && macrocount < MAXMACRO) {
	if(f & FFARG) macrocount -= 2;
	if(lastflag & CFINS) {	/* last command was insert -- tack on end */
	    macrocount--;
	    if(maclcur->l_size < maclcur->l_used + n) {
		if((lp = lallocx(maclcur->l_used + n)) == NULL)
		    return FALSE;
		lp->l_fp = maclcur->l_fp;
		lp->l_bp = maclcur->l_bp;
		lp->l_fp->l_bp = lp->l_bp->l_fp = lp;
		bcopy(maclcur->l_text, lp->l_text, maclcur->l_used);
		for(count = maclcur->l_used; count < lp->l_used; count++)
		    lp->l_text[count] = c;
		free((char *)maclcur);
		maclcur = lp;
	    } else {
		maclcur->l_used += n;
		for(count = maclcur->l_used-n; count < maclcur->l_used; count++)
		    maclcur->l_text[count] = c;
	    }
	} else {
	    macro[macrocount-1].m_funct = insert;
	    if((lp = lallocx(n)) == NULL) return FALSE;
	    lp->l_bp = maclcur;
	    lp->l_fp = maclcur->l_fp;
	    maclcur->l_fp = lp;
	    maclcur = lp;
	    for(count = 0; count < n; count++)
		lp->l_text[count] = c;
	}
	thisflag |= CFINS;
    }
#endif
    if(c == '\n') {
	do {
	    count = lnewline();
	} while (--n && count==TRUE);
	return count;
    }
    if(curbp->b_flag & BFOVERWRITE) {		/* Overwrite mode	*/
	lchange(WFEDIT);
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
	lkanji2nd = FALSE;
#endif	/* KANJI */
	while(curwp->w_doto < llength(curwp->w_dotp) && n--) {
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
	    if (lkanji2nd) {
		lkanji2nd = FALSE;
	    } else if (ISKANJI(lgetc(curwp->w_dotp, curwp->w_doto))) {
		lkanji2nd = TRUE;
	    }
#endif	/* KANJI */
	    lputc(curwp->w_dotp, curwp->w_doto++, c);
	}
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
	if (lkanji2nd && curwp->w_doto < llength(curwp->w_dotp)) {
	    lputc(curwp->w_dotp, curwp->w_doto, ' ');
	}
#endif	/* KANJI */
	if(n<=0) return TRUE;
    }
    return linsert(n, c);
}

/*
 * this could be implemented as a keymap with everthing defined
 * as self-insert.
 */
int
quote(f, n)
int f;
int n;
{
    register int c;

#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
    canna_off();
#endif
    key.k_count = 1;
    if((key.k_chars[0] = getkey(TRUE)) >= '0' && key.k_chars[0] <= '7') {
	key.k_chars[0] -= '0';
	if((c = getkey(TRUE)) >= '0' && c <= '7') {
	    key.k_chars[0] <<= 3;
	    key.k_chars[0] += c - '0';
	    if((c = getkey(TRUE)) >= '0' && c <= '7') {
		key.k_chars[0] <<= 3;
		key.k_chars[0] += c - '0';
	    } else ungetkey(c);
	} else ungetkey(c);
    }
#ifdef	CANNA	/* 95.10.26 by M.Suzuki	*/
    canna_on();
#endif
    return selfinsert(f, n);
}
