/*
 * Commands to toggle modes. Without an argument, toggle mode.
 * Negitive or zero argument, mode off.	 Positive argument, mode on.
 */
/* 90.01.29	Modified for Ng 1.0 by S.Yoshida */

#include "config.h"	/* 90.12.20  by S.Yoshida */
#include "def.h"
#include "kbd.h"

#ifndef	CHGMISC		/* 97.12.12 by M.Suzuki	*/
int defb_nmodes = 0;
MAPS *defb_modes[PBMODES] = {&map_table[0]};
int defb_flag = 0;
#endif

static int
changemode(f, n, mode)
int f;
int n;
char *mode;
{
    register int i;
    MAPS *m;

    if((m = name_mode(mode)) == NULL) {
	ewprintf("Can't find mode %s", mode);
	return FALSE;
    }
    if(!(f & FFARG)) {
	for(i=0; i <= curbp->b_nmodes; i++)
	    if(curbp->b_modes[i] == m) {
		n = 0;			/* mode already set */
		break;
	    }
    }
    if(n > 0) {
	for(i=0; i <= curbp->b_nmodes; i++)
	    if(curbp->b_modes[i] == m) return TRUE;	/* mode already set */
	if(curbp->b_nmodes >= PBMODES-1) {
	    ewprintf("Too many modes");
	    return FALSE;
	}
	curbp->b_modes[++(curbp->b_nmodes)] = m;
    } else {
	/* fundamental is b_modes[0] and can't be unset */
	for(i=1; i <= curbp->b_nmodes && m != curbp->b_modes[i]; i++) {}
	if(i > curbp->b_nmodes) return TRUE;		/* mode wasn't set */
	for(; i < curbp->b_nmodes; i++)
	    curbp->b_modes[i] = curbp->b_modes[i+1];
	curbp->b_nmodes--;
    }
    upmodes(curbp);
    return TRUE;
}

int
indentmode(f, n)
int f;
int n;
{
    return changemode(f, n, "indent");
}

int
fillmode(f, n)
int f;
int n;
{
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
    if(changemode(f, n, "fill") == FALSE) return FALSE;
    if(f & FFARG) {
	if(n <= 0) curbp->b_flag &= ~BFAUTOFILL;
	else curbp->b_flag |= BFAUTOFILL;
    } else curbp->b_flag ^= BFAUTOFILL;
    return TRUE;
#else	/* NOT KANJI */
    return changemode(f, n, "fill");
#endif	/* KANJI */
}

/*
 * Fake the GNU "blink-matching-paren" variable.
 */
int
blinkparen(f, n)
int f;
int n;
{
    return changemode(f, n, "blink");
}

#ifdef	NOTAB
int
notabmode(f, n)
int f;
int n;
{
    if(changemode(f, n, "notab") == FALSE) return FALSE;
    if(f & FFARG) {
	if(n <= 0) curbp->b_flag &= ~BFNOTAB;
	else curbp->b_flag |= BFNOTAB;
    } else curbp->b_flag ^= BFNOTAB;
    return TRUE;
}
#endif

#ifdef C_MODE	/* 90.07.24  by K.Takano */
int
cmode(f, n)
int f;
int n;
{
    return changemode(f, n, "C");
}
#endif

#ifdef	CRLFMODE		/* 97/12/10 by M.Suzuki	*/
int
crlfmode(f, n)
int f;
int n;
{
    if(changemode(f, n, "crlf") == FALSE) return FALSE;
    if(f & FFARG) {
	if(n <= 0) curbp->b_flag &= ~BFCRLF;
	else curbp->b_flag |= BFCRLF;
    } else curbp->b_flag ^= BFCRLF;
    return TRUE;
}
#endif

int
overwrite(f, n)
int f;
int n;
{
    if(changemode(f, n, "overwrite") == FALSE) return FALSE;
    if(f & FFARG) {
	if(n <= 0) curbp->b_flag &= ~BFOVERWRITE;
	else curbp->b_flag |= BFOVERWRITE;
    } else curbp->b_flag ^= BFOVERWRITE;
    return TRUE;
}

#ifdef	CHGMISC		/* 97.12.12 by M.Suzuki	*/

int
setmoremode(bp,f)
BUFFER* bp;
int f;
{
#ifdef	READONLY	/* 98.05.19 by M.Suzuki	*/
	int ret;
	BUFFER* save;

	save = curbp;
	curbp = bp;
	if( f == 1 ){
		bp->b_flag |= BFRONLY;
		ret = changemode(FFARG, 1, "readonly");
	} else
	if( f == 0 ){
		bp->b_flag &= ~BFRONLY;
		ret = changemode(FFARG, 0, "readonly");
	} else {
		bp->b_flag ^= BFRONLY;
		ret = changemode(0, 1, "readonly");
	}
	curbp = save;
	return ret;
#endif	/* READONLY */
}

int
moremode(f, n)
int f;
int n;
{
#ifdef	READONLY	/* 98.05.19 by M.Suzuki	*/
    curbp->b_flag ^= BFRONLY;
#endif	/* READONLY */
	return changemode(0,n,"readonly");
}

#ifndef NO_MACRO
#include	"macro.h"
#endif

static MODE_LIST* modeList = NULL;		/* mode list top	*/
static MODE_LIST* modeLast = NULL;		/* mode list top	*/
MODE_LIST defb = {		/* default mode	*/
	NULL,
	"*",
	0,
	{&map_table[0]},
	0,
#ifdef	TABWIDTH
	4
#endif
};

static MODE_LIST*
SearchAutoMode(str)
char* str;
{
	MODE_LIST* now;

	for(now=modeList;now;now=now->next){
		if( strcmp(str,now->name) == 0 ){
			return now;
		}
	}	/* end of for	*/
	return &defb;
}


static MODE_LIST*
CreateAutoMode(str)
char* str;
{
	MODE_LIST* now;

	if( (now=SearchAutoMode(str))!=&defb ){
		return now;
	}

	if( (now=(MODE_LIST*)calloc(1,sizeof(MODE_LIST)))==NULL ){
		return NULL;
	}
	memcpy(now,&defb,sizeof(MODE_LIST));
	now->name = (char*)strdup(str);
	if( modeList == NULL ){
		modeList = modeLast = now;
	} else {
		modeLast->next = now;
		modeLast = now;
	}
	return now;
}

static int
WildCmp( wstr, str )
unsigned char* wstr;
unsigned char* str;
{
	for(;;str++){
		if( !(*wstr | *str) ){						return TRUE;	}
		if( (*wstr == '\0') || (*str == '\0') ){	return FALSE;	}
		if( (*wstr == *str) || (*wstr == '?') ){
			wstr++;
			continue;
		}
		if( *wstr != '*' ){			return FALSE;	}
		if( !wstr[1] ){				return TRUE;	}
		if( WildCmp(wstr+1,str) ){	return TRUE;	}
	}	/* end of for	*/
}

MODE_LIST*
GetAutoMode(str)
char* str;
{
#if	0
#ifdef	REGEX
	MODE_LIST* now;

	for(now=modeList;now;now=now->next){
		if( re_strcmp(now->name,str) == 0 ){	/* It's very heavy.	*/
			return now;
		}
	}	/* end of for	*/
	return &defb;
#else
	return SearchAutoMode(str);		/* It's very pritty.	*/
#endif
#else
	MODE_LIST* now;

	for(now=modeList;now;now=now->next){
		if( WildCmp(now->name,str) == TRUE ){
			return now;
		}
	}	/* end of for	*/
	return &defb;
#endif
}

static int
_set_default_mode(f, n, ml)
int f, n;
MODE_LIST* ml;
{
	register int i;
	register MAPS *m;
	char mode[32];

	if(eread("Set mode: ", mode, 32, EFNEW) != TRUE)
		return ABORT;
	if((m = name_mode(mode)) == NULL) {
		if( strcmp(mode,"tab-width") == 0 ){
			if(eread("Tab width: ", mode, 32, EFNEW) != TRUE)
			  return ABORT;
			ml->tab_width = atoi(mode);
			ml->tab_width = ml->tab_width?ml->tab_width:1;
			return TRUE;
		}
		ewprintf("can't find mode %s", mode);
		return FALSE;
    }
    if(!(f & FFARG)) {
		for(i=0; i <= ml->nmodes; i++)
			if(ml->modes[i] == m) {
				n = 0;			/* mode already set */
				break;
			}
    }
    if(n > 0) {
		for(i=0; i <= ml->nmodes; i++)
			if(ml->modes[i] == m) return TRUE;	/* mode already set */
		if(ml->nmodes >= PBMODES-1) {
			ewprintf("Too many modes");
			return FALSE;
		}
		ml->modes[++ml->nmodes] = m;
    } else {
		/* fundamental is defb_modes[0] and can't be unset */
		for(i=1; i <= ml->nmodes && m != ml->modes[i]; i++) {}
		if(i > ml->nmodes) return TRUE;		/* mode wasn't set */
		for(; i < ml->nmodes; i++)
			ml->modes[i] = ml->modes[i+1];
		ml->nmodes--;
	}
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
	if(strcmp(mode, "fill")==0)
		if(n<=0) ml->flag &= ~BFAUTOFILL;
	else ml->flag |= BFAUTOFILL;
#endif	/* KANJI */
	if(strcmp(mode, "overwrite")==0)
		if(n<=0) ml->flag &= ~BFOVERWRITE;
	else ml->flag |= BFOVERWRITE;
#ifdef NOTAB
	if(strcmp(mode, "notab")==0)
		if(n<=0) ml->flag &= ~BFNOTAB;
	else ml->flag |= BFNOTAB;
#endif
#ifdef	CRLFMODE		/* 97/12/10 by M.Suzuki	*/
	if(strcmp(mode, "crlf")==0)
		if(n<=0) ml->flag &= ~BFCRLF;
	else ml->flag |= BFCRLF;
#endif
#ifdef	READONLY
	if( strcmp(mode,"readonly") == 0 )
		if(n<=0) ml->flag &= ~BFRONLY;
	else ml->flag |= BFRONLY;
#endif
    return TRUE;
}

int
set_default_mode(f, n)
int f, n;
{
	int ret;
	LINE* top;

	top = maclcur->l_bp;
	do {
		ret = _set_default_mode(f,n,&defb);
#ifndef	NO_MACRO
	} while( inmacro && maclcur!=top && ret == TRUE );
#else
	} while( maclcur!=top && ret == TRUE );
#endif
	return ret;
}

int
set_auto_mode(f, n)
int f;
int n;
{
	int ret;
	LINE* top;
    char fname[32];
	MODE_LIST* now;

#ifndef NO_MACRO
	if(macrodef) {
	/* keystrokes arn't collected.	Not hard, but pretty useless */
	/* would not work for function keys in any case */
	    ewprintf("Can't setup auto mode in macro");
	    return FALSE;
	}
#endif

	top = maclcur->l_bp;
    if(eread("File mask: ", fname, 32, EFNEW) != TRUE)
    	return ABORT;
	now = CreateAutoMode(fname);

	do {
		ret = _set_default_mode(f,n,now);
#ifndef	NO_MACRO
	} while( inmacro && maclcur!=top && ret == TRUE );
#else
	} while( maclcur!=top && ret == TRUE );
#endif
	return ret;
}

static VOID
list_auto_mode_line(line,now)
char* line;
MODE_LIST* now;
{
	char buff[40];
	int i;

	sprintf(buff,"\"%s\"",now->name);
	sprintf(line,"    %-20s ",buff);
	for(i=1;i<PBMODES;i++){
		if( now->modes[i] ){
			sprintf(buff,"\'%s ",now->modes[i]->p_name);
			strcat(line,buff);
		}
	}	/* end of for	*/
	sprintf(buff,"\'(tab-width %d)",now->tab_width);
	strcat(line,buff);
}

int
list_auto_mode(f, n)
int f;
int n;
{
	register BUFFER	*bp;
	register WINDOW	*wp;
	char	 line[256];
	MODE_LIST* now;

	if ((bp = bfind("*Auto Mode List*", "")) == NULL) return FALSE;
	bp->b_flag &= ~BFCHG;		/* Blow away old.	*/
 	if (bclear(bp) != TRUE) return FALSE;

	if (addline(bp, "* List of Auto mode *") == FALSE) return FALSE;
	if (addline(bp, "") == FALSE) return FALSE;

	for(now=modeList;now;now=now->next){
		list_auto_mode_line(line,now);
		if (addline(bp, line) == FALSE) return FALSE;
	}	/* end of for	*/
	list_auto_mode_line(line,&defb);
	if (addline(bp, line) == FALSE) return FALSE;

	if ((wp = popbuf(bp)) == NULL)	return FALSE;
	bp->b_dotp = lforw(bp->b_linep); /* put dot at beginning of buffer */
	bp->b_doto = 0;
	wp->w_dotp = bp->b_dotp;	/* fix up if window already on screen */
	wp->w_doto = bp->b_doto;
	return TRUE;
}

#else

int
set_default_mode(f, n)
int f;
int n;
{
    register int i;
    register MAPS *m;
    char mode[32];

    if(eread("Set Default Mode: ", mode, 32, EFNEW) != TRUE)
    	return ABORT;
    if((m = name_mode(mode)) == NULL) {
    	ewprintf("can't find mode %s", mode);
	return FALSE;
    }
    if(!(f & FFARG)) {
	for(i=0; i <= defb_nmodes; i++)
	    if(defb_modes[i] == m) {
		n = 0;			/* mode already set */
		break;
	    }
    }
    if(n > 0) {
	for(i=0; i <= defb_nmodes; i++)
	    if(defb_modes[i] == m) return TRUE;	/* mode already set */
	if(defb_nmodes >= PBMODES-1) {
	    ewprintf("Too many modes");
	    return FALSE;
	}
	defb_modes[++defb_nmodes] = m;
    } else {
	/* fundamental is defb_modes[0] and can't be unset */
	for(i=1; i <= defb_nmodes && m != defb_modes[i]; i++) {}
	if(i > defb_nmodes) return TRUE;		/* mode wasn't set */
	for(; i < defb_nmodes; i++)
	    defb_modes[i] = defb_modes[i+1];
	defb_nmodes--;
    }
#ifdef	KANJI	/* 90.01.29  by S.Yoshida */
    if(strcmp(mode, "fill")==0)
    	if(n<=0) defb_flag &= ~BFAUTOFILL;
	else defb_flag |= BFAUTOFILL;
#endif	/* KANJI */
    if(strcmp(mode, "overwrite")==0)
    	if(n<=0) defb_flag &= ~BFOVERWRITE;
	else defb_flag |= BFOVERWRITE;
#ifdef NOTAB
    if(strcmp(mode, "notab")==0)
    	if(n<=0) defb_flag &= ~BFNOTAB;
	else defb_flag |= BFNOTAB;
#endif
#ifdef	CRLFMODE		/* 97/12/10 by M.Suzuki	*/
    if(strcmp(mode, "crlf")==0)
    	if(n<=0) defb_flag &= ~BFCRLF;
	else defb_flag |= BFCRLF;
#endif
    return TRUE;
}
#endif	/* CHGMISC	*/

