/*
 * fepctrl.c 1.2 1990/11/25	Public Domain.
 *
 * General purpose Japanese FEP control routines for MS-DOS.
 * Written by Junn Ohta <ohta@src.ricoh.co.jp>
 *
 *	int fep_init()
 *		checks FEP and turn it off, returns FEP type.
 *	void fep_term()
 *		restore the status of FEP saved by fep_init().
 *	void fep_on()
 *		restore the status of FEP saved by fep_off().
 *	void fep_off()
 *		save the status of FEP and turn it off.
 *	void fep_force_on()
 *		turn FEP on by its default "on" status.
 *	void fep_force_off()
 *		don't save the status of FEP and turn it off.
 */

#define	NG
#ifdef	NG
#include "config.h"
#endif

#ifdef	FEPCTRL

#ifndef __TURBOC__
/*
 * use assembler routines to support ATOK7/FIXER direct control
 */
#define	USE_FEPCSUB
#endif

#define DO_MSKANJI	/* support MS-KANJI application interface */
#define DO_GENERIC	/* support platforms other than PC-9801 */

#include <dos.h>
#include "fepctrl.h"

/*
 * default "on" status of FEP (used only in fep_force_on())
 */
static int fepon[NFEPS] = {
/* FEP_NONE	*/  0,
/* FEP_PC98A	*/  1,	/* 0=off, 1=on */
/* FEP_PC98B	*/  1,	/* 0=off, 1=on */
/* FEP_MSKANJI	*/  2,	/* 0=off, 1=system, 2=echo */
/* FEP_VJE	*/  2,	/* 0=off, 1=system, 2=echo */
/* FEP_ATOK6	*/  1,	/* 0=off, 1=roma, 2=kana, 3=alpha, 4=sym, 5=code */
/* FEP_ATOK7	*/  1,	/* 0=off, 1=on */
/* FEP_MTTK86	*/  1,  /* 0=off, 1=on */
/* FEP_MTTK2	*/  1,  /* 0=off, 1=on */
/* FEP_KATANA	*/  1,  /* 0=off, 1=echo, 2=system */
/* FEP_FIXER	*/  1,  /* 0=off, 1=echo, 2=system */
/* FEP_EGB2	*/  2,  /* 0=off, 1=system, 2=echo */
/* FEP_EGB3	*/  2,  /* 0=off, 1=system, 2=echo */
/* FEP_WXP	*/  1,  /* 0=off, 1=echo, 2=system */
/* FEP_MGR2	*/  1,  /* 0=off, 1=on */
/* FEP_NEC	*/  1,  /* 0=off, 1=on */
/* FEP_DFJ	*/  1,  /* 0=off, 1=on */
/* FEP_KAZE	*/  1,	/* 0=off, 1=on */
/* FEP_OAK   */  0xd1,	/* 0r010100=off, 1r010001=on (r=romaji) */
/* FEP_B16   */  0x80,	/* 00000000=off, 10000000=on */
/* FEP_WXPJ	*/  1	/* 0=off, 1=echo, 2=system */
};

static int fep = FEP_NONE;
static int oldmode = 0;
static int keepmode = 0;

static void fep_open(), fep_close();
static int fep_find(), fep_mode();
static int check_dev(), check_vec();
static int msknj_find();
static int msknj_get_mode(), msknj_set_mode();
static int msknj_kkfunc();

int
fep_init()
{
	if (fep == FEP_NONE)
		fep = fep_find();
	fep_open();
	oldmode = keepmode = fep_mode(0);
	return fep;
}

void
fep_term()
{
	fep_mode(oldmode);
	fep_close();
}

void
fep_on()
{
	fep_mode(keepmode);
}

void
fep_off()
{
	keepmode = fep_mode(0);
}

void
fep_force_on()
{
	fep_mode(fepon[fep]);
}

void
fep_force_off()
{
	fep_mode(0);
}

/*--------------------------------------------------------------------*/
#ifdef TEST

static char *fepname[NFEPS] = {
/* FEP_NONE	*/  "(none)",
/* FEP_PC98A	*/  "PC9801A",
/* FEP_PC98B	*/  "PC9801B",
/* FEP_MSKANJI	*/  "MSKANJI",
/* FEP_VJE	*/  "VJE",
/* FEP_ATOK6	*/  "ATOK6",
/* FEP_ATOK7	*/  "ATOK7",
/* FEP_MTTK86	*/  "MTTK86",
/* FEP_MTTK2	*/  "MTTK2",
/* FEP_KATANA	*/  "KATANA",
/* FEP_FIXER	*/  "FIXER",
/* FEP_EGB2	*/  "EGB2",
/* FEP_EGB3	*/  "EGB3",
/* FEP_WXP	*/  "WXP",
/* FEP_MGR2	*/  "MGR2",
/* FEP_NEC	*/  "NEC",
/* FEP_DFJ	*/  "DFJ",
/* FEP_KAZE	*/  "KAZE",
/* FEP_OAK	*/  "OAK",
/* FEP_B16	*/  "B16",
/* FEP_WXPJ	*/  "WXPJ"
};

#define	getch()		(bdos(8, 0, 0) & 0xff)
#define	putch(c)	bdos(2, (int)(c), 0)

static void
putstr(s)
char *s;
{
	while (*s)
		putch(*s++);
}

static void
echoline(s)
char *s;
{
	int c;
	putstr(s);
	while ((c = getch()) != '\r' && c != '\n')
		putch(c);
	putstr("\r\n");
}

main()
{
	putstr("fep = ");
	putstr(fepname[fep_init()]);
	putstr("\r\n");
	putstr("enter 4 lines of text\r\n");
	fep_force_on();
	echoline(" on: ");
	fep_off();
	echoline("off: ");
	fep_on();
	echoline(" on: ");
	fep_force_off();
	echoline("off: ");
	fep_term();
	exit(0);
}

#ifdef __TURBOC__
_setargv(){}
_setenvp(){}
exit(c){_exit(c);}
#endif

#endif /* TEST */
/*--------------------------------------------------------------------*/

#ifdef LDATA
#undef LDATA
#endif
#if defined(M_I86CM) || defined(M_I86LM) || defined(M_I86HM)
#define	LDATA
#endif
#if defined(__COMPACT__) || defined(__LARGE__) || defined(__HUGE__)
#define	LDATA
#endif

#define	uns	unsigned

typedef struct {
	int	ftype;		/* fep type */
	char	*dname;		/* device name */
} FEPDEV;

/*
 * device names supplied by user for FEP residence check
 */
static FEPDEV fepudev[] = {
	{ FEP_PC98A,	"FP$PC98A" },
	{ FEP_PC98B,	"FP$PC98B" },
	{ FEP_PC98B,    "FP$DUMMY" },	/* for backward compatibility */
	{ FEP_PC98B,    "FP$PC98"  },	/* for backward compatibility */
	{ FEP_VJE,      "FP$VJE"   },
	{ FEP_ATOK6,    "FP$ATOK6" },
	{ FEP_MTTK86,   "FP$MTTK"  },
	{ FEP_KATANA,   "FP$KTN"   },
	{ FEP_EGB2,     "FP$EGB2"  },
	{ FEP_EGB3,     "FP$EGB3"  },
	{ FEP_WXP,      "FP$WXP"   },
	{ FEP_MGR2,     "FP$MGR2"  },
	{ FEP_NEC,      "FP$NEC"   },
	{ FEP_DFJ,      "FP$DFJ"   },
	{ FEP_KAZE,	"FP$KAZE"  },
#ifdef DO_GENERIC
	{ FEP_OAK,	"FP$OAK"   },
#endif
	{ FEP_NONE, }
};

/*
 * real device names built in by FEPs
 */
static FEPDEV fepdev[] = {
	{ FEP_VJE,	"AS$VJE"   },
	{ FEP_ATOK6,	"ATOKF%%%" },
	{ FEP_MGR2,	"#MGR#REX" },
	{ FEP_NEC,	"$AID#NEC" },	/* AI Kana-Kanji Henkan on DOS3.3C */
	{ FEP_NEC,	"$AIC#NEC" },	/* AI Kana-Kanji Henkan */
	{ FEP_NEC,	"$ATC#NEC" },	/* Chikuji Kana-Kanji Henkan */
	{ FEP_NONE, }
};

typedef struct {
	int	ftype;		/* fep type */
	uns int	inum;		/* dointr number */
	uns int	offset;		/* offset of id string */
	uns int relative;	/* offset is relative? */
	char	*idstr;		/* id string */
} FEPINT;

/*
 * real ID string placed in the memory
 */
static FEPINT fepint[] = {
	{ FEP_MTTK86,	0x70, 0x12, 0, "MTTK86"   },
	{ FEP_MTTK2,	0x70, 0x12, 0, "MTTK2"    },
	{ FEP_EGB2,	0x70, 0x28, 0, "ERGOA"    },
	{ FEP_EGB3,	0x70, 0x28, 0, "ERGO"     },
	{ FEP_WXP,	0x70, 0x22, 0, "WX      " },
	{ FEP_DFJ,	0x6f, 0x02, 0, "DFJ"      },
	{ FEP_ATOK6,	0x6f, 0x03, 1, "ATOK5020" },
	{ FEP_ATOK6,	0x6f, 0x1d, 0, "ATOK5020" }, /* backward compat. */
/*	{ FEP_VJE,	0x70, 0x12, 0, "VJE-"     }, */
/*	{ FEP_MGR2,	0x70, 0x0a, 0, "#MGR#REX" }, */
/*	{ FEP_NEC,	0xdc, 0x0a, 0, "$AID#NEC" }, */
/*	{ FEP_NEC,	0xdc, 0x0a, 0, "$ATC#NEC" }, */
#ifdef DO_GENERIC
	{ FEP_OAK,	0xed, 0x12, 0, "OAK000"   },
	{ FEP_B16,	0x16, 0x02, 1, "K,YUGEV"  },
	{ FEP_WXPJ,	0x6c, 0x22, 0, "WX      " },
#endif
	{ FEP_NONE, }
};

/*
 * FIXER ioctl data packet
 */
typedef struct {
	char	name[8];
	int	version;
	void	(far *v3func)();	
	int	v3vec;
	void	(far *v4func)();
	int	v4vec;
} FIXIDP;

#ifdef __TURBOC__
static void (far *fixer_entry)();
static void (far *atok7_entry)();
#else
#ifdef USE_FEPCSUB
void (far *fixer_entry)();
void (far *atok7_entry)();
#endif
#endif
static int fixer_vec = -1;

#define GET_IOC_DATA	0x4400	/* Get IOCTL Data		*/
#define	RCV_IOC_CHAR	0x4402	/* Receive IOCTL Character	*/
#define	GET_CBRK	0x3300	/* Get CBREAK (CTRL-C Check)	*/
#define	SET_CBRK	0x3301	/* Set CBREAK (CTRL-C Check)	*/
#define	GET_VECTOR	0x35	/* Get Interrupt Vector		*/
#define	OPEN_HANDLE	0x3d	/* Open Handle			*/
#define	CLOSE_HANDLE	0x3e	/* Close Handle			*/
#define DEVICE		0x80	/* handle is not a file		*/

#define	MK_FAR(seg, off) (uns char far *)((uns long)(seg) << 16 | (uns)(off))
#define	FAR_SEG(addr)	 ((unsigned)((unsigned long)(addr) >> 16))
#define	FAR_OFF(addr)	 ((unsigned)(addr))

static union REGS regs;
static struct SREGS sregs;
static unsigned char *p;
static unsigned char far *fp;

#ifndef __TURBOC__	/* Microsoft C */
#define	disable()	_disable()
#define	enable()	_enable()
#endif

static int
fep_find()
{
	int i;
	FEPDEV *fv;
	FEPINT *fi;
	FIXIDP fixer;

	/*
	 * determine fep type by its USER SUPPLIED device name
	 */
	for (fv = fepudev; fv->ftype != FEP_NONE; fv++)
		if (check_dev(fv->dname))
			return fv->ftype;

	/*
	 * loaded into bank memory by MELCO's FEP loader
	 */
	if (check_dev("FPLD$10")) {
		if (check_dev("AS$VJE"))
			return FEP_VJE;		/* don't use MS-KANJI API */
		if (check_vec(0x70, 0x12, 0, "MTTK2"))
			return FEP_MTTK86;	/* don't use INT 70H I/F */
		if (check_dev("$AID#NEC"))
			return FEP_NEC;		/* don't use MS-KANJI API */
	}

#ifdef DO_MSKANJI
	/*
	 * MS-KANJI FEP
	 */
	if (check_dev("MS$KANJI") && msknj_find())
		return FEP_MSKANJI;
#endif

	/*
	 * determine fep type by its device name
	 */
	for (fv = fepdev; fv->ftype != FEP_NONE; fv++)
		if (check_dev(fv->dname))
			return fv->ftype;

	/*
	 * determine fep type by its id string
	 */
	for (fi = fepint; fi->ftype != FEP_NONE; fi++)
		if (check_vec(fi->inum, fi->offset, fi->relative, fi->idstr))
			return fi->ftype;

#if defined(__TURBOC__) || defined(USE_FEPCSUB)
	/*
	 * special case: ATOK7 function call
	 */
	regs.h.ah = GET_VECTOR;
	regs.h.al = 0x6f;
	intdosx(&regs, &regs, &sregs);
	fp = MK_FAR(sregs.es, regs.x.bx + 0x30);
	for (p = "ATOK7"; *p; p++)
		if (*p != *fp++)
			break;
	if (!*p) {
		atok7_entry = (void far *)MK_FAR(sregs.es, regs.x.bx + 0x20);
		return FEP_ATOK7;
	}
#endif

	/*
	 * special case: FIXER3/4 application interface
	 */
	regs.x.ax = GET_IOC_DATA;
	regs.x.bx = 0;	/* stdin */
	intdos(&regs, &regs);
	if (regs.x.dx & DEVICE) {
		regs.x.ax = RCV_IOC_CHAR;
		regs.x.bx = 0;	/* stdin */
		regs.x.cx = sizeof(fixer);
#ifdef LDATA
		regs.x.dx = FAR_OFF(&fixer);
		sregs.ds = FAR_SEG(&fixer);
		intdosx(&regs, &regs, &sregs);
#else
		regs.x.dx = (unsigned int)&fixer;
		intdos(&regs, &regs);
#endif
		i = regs.x.ax;
		if (i >= 8 && !strncmp(fixer.name, "FIXER   ", 8)) {
			if (i >= 16 && fixer.v3vec != -1) {
				fixer_vec = fixer.v3vec;
				return FEP_FIXER;
			}
			if (i >= 22 && fixer.v4vec != -1) {
				fixer_vec = fixer.v4vec;
				return FEP_FIXER;
			}
#if defined(__TURBOC__) || defined(USE_FEPCSUB)
			if (check_dev("FPLD$10") || check_dev("$BNK#DRV"))
				return FEP_PC98B;
			if (i >= 16 && fixer.v3func != (void far *)0) {
				fixer_entry = fixer.v3func;
				return FEP_FIXER;
			}
			if (i >= 22 && fixer.v4func != (void far *)0) {
				fixer_entry = fixer.v4func;
				return FEP_FIXER;
			}
#endif
			return FEP_PC98B;
		}
	}

	/*
	 * special case: Katana/ACE software dointr interface
	 */
	if (check_vec(0x48, 0x0a, 0, "CON     ") &&
	    check_vec(0x49, 0x0a, 0, "CON     ") &&
	    check_vec(0x4a, 0x0a, 0, "CON     ")) {
		/*
		 * This may not be a good way to check Katana/ACE.
		 * However, console status report (sending "<ESC>@"
		 * returns "$") is not reliable here, because we
		 * may have characters in the keyboard buffer.
		 */
		return FEP_KATANA;
	}

	return FEP_NONE;
}

static int
dopen(dname)
char *dname;
{
#ifdef	NG	/* 91.01.18  by Junn Ohta */
	static char buf[10] = "\\";
	char *strcpy();

	strcpy(buf+1, dname);
	dname = buf;
#endif	/* NG */

	regs.h.ah = OPEN_HANDLE;
	regs.h.al = 0;
#ifdef LDATA
	regs.x.dx = FAR_OFF(dname);
	sregs.ds = FAR_SEG(dname);
	intdosx(&regs, &regs, &sregs);
#else
	regs.x.dx = (unsigned int)dname;
	intdos(&regs, &regs);
#endif
	if (regs.x.cflag)
		return -1;
	return regs.x.ax;
}

static void
dclose(fd)
int fd;
{
	regs.h.ah = CLOSE_HANDLE;
	regs.x.bx = fd;
	intdos(&regs, &regs);
}

static int
check_dev(dname)
char *dname;
{
	int fd, dc;

	if ((fd = dopen(dname)) < 0)
		return 0;

	regs.x.ax = GET_IOC_DATA;
	regs.x.bx = fd;
	intdos(&regs, &regs);
	dc = regs.x.dx;
	dclose(fd);
	return (dc & DEVICE)? 1: 0;
}

static int
check_vec(inum, offset, relative, idstr)
unsigned int inum, relative, offset;
char *idstr;
{
	/*
	 * get dointr vector
	 */
	regs.h.ah = GET_VECTOR;
	regs.h.al = inum;
	intdosx(&regs, &regs, &sregs);
	/*
	 * check id string
	 */
	if (relative)
		offset += regs.x.bx;
	fp = MK_FAR(sregs.es, offset);
	for (p = idstr; *p; p++)
		if (*p != *fp++)
			return 0;
	return 1;
}

static int
call_fixer(func, val)
int func, val;
{
#ifdef __TURBOC__
	if (fixer_vec != -1) {
		_AX = func;
		_DL = (unsigned char)val;
		geninterrupt(fixer_vec);
		return (int)_DL;
	}
	_AX = func;
	_DL = (unsigned char)val;
	(*fixer_entry)();
	return (int)_DL;
#else /* Microsoft C */
#ifdef USE_FEPCSUB
	extern int fixer_sub();
	if (fixer_vec != -1) {
		regs.x.ax = func;
		regs.h.dl = (unsigned char)val;
		int86(fixer_vec, &regs, &regs);
		return regs.h.dl;
	}
	return fixer_sub(func, val);
#else
	regs.x.ax = func;
	regs.h.dl = (unsigned char)val;
	int86(fixer_vec, &regs, &regs);
	return regs.h.dl;
#endif
#endif
}

static int
call_atok7(func, val)
int func, val;
{
#ifdef __TURBOC__
	int	packet[32];

	*packet = val;
	_ES = FP_SEG(packet);
	_BX = FP_OFF(packet);
	_AX = func;
	(*atok7_entry)();
	return *packet;
#else
#ifdef USE_FEPCSUB
	extern int atok7_sub();
	return atok7_sub(func, val);
#else
	return 0;
#endif
#endif
}

/*
 * PC-9801 CRT/KBD BIOS, VRAM
 */
#define	CRTWORK   ((unsigned char far *)MK_FAR(0x0000, 0x053d))
#define	CRTLINES  ((unsigned char far *)MK_FAR(0x0000, 0x0713))
#define	VRAM	  ((unsigned char far *)MK_FAR(0xa000, 0x0000))
#define	KBDBUF    ((unsigned short far *)MK_FAR(0x0000, 0x0502))
#define	KBDCNT    ((unsigned short far *)MK_FAR(0x0000, 0x0528))
#define	KBDTAIL   ((unsigned short near *(far *))MK_FAR(0x0000, 0x0526))
#define	KBDMAX    16		/* max # of keys in KBDBUF */
#define	CTRLXFER  0xb500	/* CTRL-XFER key scan code */

/*
 * Look into CRT BIOS work area and check FEP status
 */
static int
pc98_crtmode()
{
	return (*CRTWORK == 2)? 1: 0;
}

/*
 * Look into VRAM bottom right corner and check FEP status
 */
static int
pc98_vrammode()
{
	return (VRAM[*CRTLINES? 0x0f9d: 0x0c7d] > '!')? 1: 0;
}

/*
 * Set key code directly into PC-9801 BIOS key code buffer
 */
static void
pc98_setkey(keycode)
unsigned int keycode;
{
	disable();
	*(unsigned short far *)(MK_FAR(0x0000, *KBDTAIL)) = keycode;
	(*KBDCNT)++;
	(*KBDTAIL)++;
	if (*KBDTAIL >= (unsigned short near *)FAR_OFF(&KBDBUF[KBDMAX]))
		*KBDTAIL = (unsigned short near *)FAR_OFF(KBDBUF);
	enable();
}

static void
pc98_flushkey(keycode)
unsigned int keycode;
{
	while (bdos(6, 0xff, 0) & 0xff) /* void */;
	pc98_setkey(keycode);
	while (bdos(6, 0xff, 0) & 0xff) /* void */;
}

static int
set_cbrk(newcbrk)
int newcbrk;
{
	int curcbrk;

	/*
	 * get current cbrk value
	 */
	regs.x.ax = GET_CBRK;
	intdos(&regs, &regs);
	curcbrk = regs.h.dl;
	/*
	 * set new cbrk value
	 */
	regs.x.ax = SET_CBRK;
	regs.h.dl = newcbrk;
	intdos(&regs, &regs);
	return curcbrk;
}

static int
dointr(vec, ah, al)
int vec, ah, al;
{
	regs.h.ah = ah;
	regs.h.al = al;
	int86(vec, &regs, &regs);
	return regs.h.al;
}

/*
 * NEC extended function call (int DCh)
 */
static void
donec(cl)
int cl;
{
	regs.x.ax = 0;
	regs.h.cl = cl;
	int86(0xdc ,&regs, &regs);
}

static void
fep_open()
{
	if (fep == FEP_NEC)
		donec(224);
}

static void
fep_close()
{
	if (fep == FEP_NEC)
		donec(225);
}

static int
fep_mode(newmode)
int newmode;
{
	int curmode;

	switch (fep) {
#ifdef DO_MSKANJI
	case FEP_MSKANJI:
		/* mode: 0 = off, 1 = system, 2 = echo */
		curmode = msknj_get_mode();
		if (newmode != curmode)
			msknj_set_mode(newmode);
		return curmode;
#endif
	case FEP_VJE:
	case FEP_EGB2:
	case FEP_EGB3:
		/* mode: 0 = off, 1 = system, 2 = echo */
		curmode = dointr(0x70, 0, 0);
		if (newmode != curmode) {
			dointr(0x70, 1, newmode);
			if (fep == FEP_EGB2)
				dointr(0x70, 12, 0); /* unlock mode key */
		}
		return curmode;
	case FEP_ATOK6:
		/* mode: 0=off, 1=roma, 2=kana, 3=alpha, 4=sym, 5=code */
		curmode = dointr(0x6f, 102, 0);
		if (newmode != curmode)
			dointr(0x6f, (!newmode? 11: newmode), 0);
		return curmode;
	case FEP_ATOK7:
		/* mode: 0 = off, 1 = on */
		curmode = call_atok7(3, -1)? 1: 0;
		if (newmode != curmode)
			call_atok7(newmode? 1: 2, 0);
		return curmode;
	case FEP_MTTK2:
		/* mode: 0 = off, 1 = on */
		regs.h.ah = 21;
		int86x(0x70, &regs, &regs, &sregs);
		fp = MK_FAR(sregs.es, regs.x.bx);
		curmode = fp[5];
		if (newmode != curmode)
			dointr(0x70, 0, newmode);
		return curmode;
	case FEP_FIXER:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = call_fixer(0x0901, 0);
		if (newmode != curmode)
			call_fixer(0x0900, newmode);
		return curmode;
	case FEP_KATANA:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = dointr(0x48, 2, 0);
		if (newmode != curmode) {
			int old = set_cbrk(0);
			dointr(0x48, 3, newmode);
			set_cbrk(old);
		}
		return curmode;
	case FEP_WXP:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = dointr(0x70, 0x4c, 0);
		if (newmode == 0 && curmode != 0) {
			while (dointr(0x70, 0x4c, 0))
				pc98_flushkey(CTRLXFER);
		} else if (newmode != 0 && curmode == 0) {
			dointr(0x70, 0x5c, newmode);
			dointr(0x70, 0x52, 1);
		}
		return curmode;
	case FEP_MGR2:
		/* mode: 0 = off, 1 = on */
		regs.h.ah = 0;
		int86(0x70, &regs, &regs);
		curmode = (regs.h.dl & 0x01);
		if (newmode != curmode)
			dointr(0x46, 0, newmode);
		return curmode;
	case FEP_KAZE:
		/* mode: 0 = off, 1 = on */
		curmode = pc98_vrammode();
		if (newmode != curmode)
			pc98_setkey(CTRLXFER);
		return curmode;
	case FEP_NEC:
		/* mode: 0 = off, 1 = on */
		curmode = pc98_crtmode();
		if (newmode != curmode)
			donec(newmode? 239: 240);
		return curmode;
	case FEP_PC98A:
	case FEP_MTTK86:
	case FEP_DFJ:
		/* mode: 0 = off, 1 = on */
		curmode = pc98_crtmode();
		while (pc98_crtmode() != newmode)
			pc98_flushkey(CTRLXFER);
		return curmode;
	case FEP_PC98B:
		/* mode: 0 = off, 1 = on */
		curmode = pc98_crtmode();
		if (newmode != curmode)
			pc98_setkey(CTRLXFER);
		return curmode;
#ifdef DO_GENERIC
	case FEP_OAK:
		/*
		 * mode: 0r010100=off, 1r010001=on (r=romaji)
		 *   bit7: size: 0 = hankaku, 1 = zenkaku
		 *   bit6: romaji entry: 0 = off, 1 = on
		 *   bit5: always 0
		 *   bit4: mode change: 0 = disabled, 1 = enabled
		 *   bit3: always 0
		 * bit2-0: char type: 001 = hiragana, 010 = katakana,
		 *	   011 = uppercase, 100 = lowercase, 101 = al
		 */
#define	OAKOFF(mode)	((mode) == 0 || ((mode) & 0x84) == 0x04)
		curmode = dointr(0xec, 3, 0);
		if (OAKOFF(newmode) && !OAKOFF(curmode))
			dointr(0xec, 2, (curmode & 0x50) | 0x04);
		else if (!OAKOFF(newmode) && OAKOFF(curmode))
			dointr(0xec, 2, newmode);
		return curmode;
	case FEP_B16:
		/* mode: 0000000 = off, 1000000 = on */
		curmode = dointr(0x16, 8, 0) & 0x80;
		if (newmode != curmode)
			dointr(0x16, 7, newmode);
		return curmode;
	case FEP_WXPJ:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = dointr(0x6c, 0x4c, 0);
		if (newmode != curmode)
			dointr(0x6c, 0, newmode);
		return curmode;
#endif
	default:
		return 0;
	}
}

#ifdef DO_MSKANJI
/*
 * MS-KANJI Application Interface access routines
 */

struct Kkname {
	short		wLevel;			/* API support level */
	char		rgchName[8];		/* name of system */
	char		rgchVersion[4];		/* version # of system */
	char		rgchRevision[4];	/* revision # of system */
	char		Reserved[14];		/* reserved */
};
struct Dataparm {
	short		wType;			/* display switch */
	short		wScan;			/* key scan code */
	short		wAscii;			/* ascii code */
	short		wShift;			/* shift key status */
	short		wExShift;		/* extended shift key status */
	short		cchResult;		/* size of confirmed string */
	char far	*lpchResult;		/* ptr to confirmed string */
	short		cchMode;		/* size of mode display str */
	char far	*lpchMode;		/* ptr to mode display str */
	char far	*lpattrMode;		/* ptr to attribute of it */
	short		cchSystem;		/* size of system string */
	char far	*lpchSystem;		/* ptr to system string */
	char far	*lpattrSystem;		/* ptr to attribute of it */
	short		ccBuf;			/* size of unconfirmed str */
	char far	*lpchBuf;		/* ptr to unconfirmed str */
	char far	*lpattrBuf;		/* ptr to attribute of it */
	short		cchBufCursor;		/* cursor position in it */
	char		Reserved[34];		/* reserved */
};
struct Funcparm {
	short		wFunc;			/* function # */
	short		wMode;			/* reference or set */
	struct Kkname far *lpKkname;		/* translation system */
	struct Dataparm far *lpDataparm;	/* data parameter */
	char		Reserved[4];		/* reserved */
};

/*
 * function # for wFunc
 */
#define	KKAsk	1
#define	KKOpen	2
#define KKClose	3
#define	KKInOut	4
#define	KKMode	5

/*
 * mode bit for wMode
 */
#define	KM_GET	0x0000
#define	KM_SET	0x8000
#define	KM_OFF	0x0001
#define	KM_ON	0x0002
#define	KM_SYS	0x0004
#define	KM_ECHO	0x0008

static int (far pascal *KKfunc)();
#ifndef __TURBOC__
static int (far *KKfunc_vje)();
#endif
static struct Funcparm Funcbuf;
static struct Kkname Kkbuf;

static int
msknj_find()
{
	int fd, n;

	/*
	 * get entry address of KKfunc()
	 */
	fd = dopen("MS$KANJI");
	regs.x.ax = RCV_IOC_CHAR;
	regs.x.bx = fd;
	regs.x.cx = 4;
#ifdef LDATA
	regs.x.dx = FAR_OFF(&KKfunc);
	sregs.ds = FAR_SEG(&KKfunc);
	intdosx(&regs, &regs, &sregs);
#else
	regs.x.dx = (unsigned int)&KKfunc;
	intdos(&regs, &regs);
#endif
	n = regs.x.ax;
	dclose(fd);
	if (n != 4)		/* ioctl failed */
		return 0;

	/*
	 * We don't use MS-KANJI API upon ATOK7 with MSATOK,
	 * because SoftVision's MSATOK doesn't satisfy the
	 * specification of KKMode function.
	 */
	if (check_vec(0x6f, 0x30, 1, "ATOK7"))
		return 0;

#ifndef __TURBOC__
	/*
	 * We should call KKfunc() as a C function instead of a
	 * pascal one, because KKfunc()s implemented in VJE-Beta
	 * and NEC AI03.00 doesn't restore the stack pointer.
	 */
	if (check_dev("AS$VJE") || check_dev("$AID#NEC"))
		KKfunc_vje = (int (far *)())KKfunc;
#endif

	/*
	 * check API support level
	 */
	memset(&Funcbuf, 0, sizeof(Funcbuf));
	Funcbuf.wFunc = KKAsk;
	Funcbuf.lpKkname = (struct Kkname far *)&Kkbuf;
	if (msknj_kkfunc((struct Funcparm far *)&Funcbuf) < 0)
		return 0;
	if (Kkbuf.wLevel < 1)	/* level too low */
		return 0;
	return 1;
}

static int
msknj_get_mode()
{
	memset(&Funcbuf, 0, sizeof(Funcbuf));
	Funcbuf.wFunc = KKMode;
	Funcbuf.wMode = KM_GET;
	msknj_kkfunc((struct Funcparm far *)&Funcbuf);
	return (Funcbuf.wMode & KM_OFF)? 0: (Funcbuf.wMode & KM_SYS)? 1: 2;
}

static int
msknj_set_mode(mode)
int mode;
{
	memset(&Funcbuf, 0, sizeof(Funcbuf));
	Funcbuf.wFunc = KKMode;
	Funcbuf.wMode = KM_SET;
	Funcbuf.wMode |= (mode == 0)? KM_OFF: KM_ON;
	Funcbuf.wMode |= (mode == 1)? KM_SYS: KM_ECHO;
	msknj_kkfunc((struct Funcparm far *)&Funcbuf);
	return 0;
}

static int
msknj_kkfunc(bufp)
struct Funcparm far *bufp;
{
#ifdef __TURBOC__
	static int sp, ret;

	sp = _SP;
	ret = (*KKfunc)((struct Funcparm far *)bufp);
	_SP = sp;
	return ret;
#else
	if (KKfunc_vje)
		return (*KKfunc_vje)((struct Funcparm far *)bufp);
	else
		return (*KKfunc)((struct Funcparm far *)bufp);
#endif
}
#endif /* DO_MSKANJI */
#endif /* FEPCTRL */
